#!/usr/bin/perl
use strict;
use warnings;
use Cwd;

#  PROGRAM JeTU# Jean Baptiste Rota (Avril 2009) Martin Vérot (janvier/février 2010)
#  le programme 
#	i) lit les coefficients dans la fonction d'onde DDCI. 
#	ii) Extrait les paramètres J, t, et U.

#Publication originale :
#Calzado, C. J. ; Cabrero, J. ; Malrieu, J. P. ; Caballol, R. The Journal of Chemical Physics 2002, 116, 3985-4000.




#######Le programme prend en arguments :
#les fonctions d'ondes DDCI données sous forme de matrice 
#  Lambda fondamental		lambda état de type "u"		lambda du singulet "g" excité
#	mu fondamental				mu état u						mu état g excité
#  nu du fondamental			nu état u						nu état g excité
#Énergie du singulet fondamental
#énergie du triplet
#énergie du singulet de type u
#énergie du singulet de type g

# normalement 
#####"Lambda fondamental">"mu fondamental"	 et les deux ont un signe opposé (avec nu nul ou négligeable)
#####"mu état excité">"lambda état excité" (avec nu nul ou petit)
#####"nu état u" avec le poids prépondérant 

#ci après un fichier d'entrée donné en exemple

#                    -0.6933831                    -0.1175236                    -0.4999537 ###PREMIERE LIGNE DU FICHIER
#                     0.6104619                    -0.1447147                    -0.6100782
#                             0                    -0.5577670                     0.1320357
#-2870.220211899988
#-2870.219684961527
#-2870.077152558950
#-2870.076413308386



	open(my $in_coeff, "<",  "$ARGV[0]") or die "Can't open intput.txt: $!";	
	open(my $out, ">",  "$ARGV[0].out") or die "Can't open output.txt: $!";	

##############################################
############FONCTIONS#########################
##############################################
my $i;my $j;my $k;
#Multiplication de matrice
sub matrix_multiply {
    my ($r_mat1,$r_mat2)=@_;
    my ($r_product);
    my ($r1,$c1)=matrix_count_rows_cols($r_mat1);
    my ($r2,$c2)=matrix_count_rows_cols($r_mat2);
    
    die "matrix 1 has $c1 columns and matrix 2 has $r2 rows>" 
        . " Cannot multiply\n" unless ($c1==$r2);
    for (my $i=0;$i<$r1;$i++) {
        for (my $j=0;$j<$c2;$j++) {
            my $sum=0;
            for (my $k=0;$k<$c1;$k++) {
                $sum+=$r_mat1->[$i][$k]*$r_mat2->[$k][$j];
            }
            $r_product->[$i][$j]=$sum;
        }
    }
    $r_product;
}


sub matrix_add {
    my ($r_mat1,$r_mat2)=@_;
    my ($r_add);
    my ($r1,$c1)=matrix_count_rows_cols($r_mat1);
    my ($r2,$c2)=matrix_count_rows_cols($r_mat2);
    
    die "matrix 1 has $c1 columns and matrix 2 has $r2 rows>" 
        . " Cannot multiply\n" unless ($c1==$c2 && $r1==$r2);
   for (my $i=0;$i<$r1;$i++) 
	{
	   for (my $j=0;$j<$c1;$j++) 
		{
			$r_add->[$i][$j]=$r_mat1->[$i][$j]+$r_mat2->[$i][$j];
   	}
   }
   $r_add;
}


sub matrix_multiply_scalar {
    my ($r_mat1,$scalar)=@_;
    my ($r_prod);
    my ($r1,$c1)=matrix_count_rows_cols($r_mat1);
   for (my $i=0;$i<$r1;$i++) 
	{
	   for (my $j=0;$j<$c1;$j++) 
		{
			$r_prod->[$i][$j]=$r_mat1->[$i][$j]*$scalar;
   	}
   }
   $r_prod;
}



sub matrix_transpose {
	my ($r_mat1)=@_;
	my ($r_transpo);
	my ($r1,$c1)=matrix_count_rows_cols($r_mat1);
	for (my $i=0;$i<$r1;$i++) 
	{
		for (my $k=0;$k<$c1;$k++) 
			{
				$r_transpo->[$i][$k]=$r_mat1->[$k][$i];
			}
	}
	$r_transpo;
}


#Calcul des dimensions de la matrice
sub matrix_count_rows_cols { 
    my ($r_mat)=@_;
    my $num_rows=@$r_mat;
    my $num_cols=@{$r_mat->[0]};
    ($num_rows,$num_cols);
}

#écrire la matrice dans le fichier de sortie
sub matrix_print
{
my $l;
my @mat=@_;
my $i=0;my $j;
for ($i=0; $i<@mat; $i++) 
{
	for ($j=0; $j<@mat; $j++) 
	{
		$l=length($mat[$i][$j]);
		print $out " "x(30-$l).$mat[$i][$j]
	}
	print $out "\n";
}
}	

#Afficher un vecteur colonne.
sub colonne_print
{
	my @mat=@_;
	my $i=0;
	for ($i=0; $i<@mat; $i++) 
		{
			print $out $mat[$i]."\n";
		}
}
	
##############################################


my $test;
	

	
#On lit coefficients et énergies	dans le fichier d'entrée	
	my @Psi_lue;
	my @Energie_lue;
	$test=0;
while (<$in_coeff>) 
{ 
	#On lit la fonction d'onde DDCI
	if($test<3 && /^\s*([\d\.\+\-]*)\s*([\d\.\+\-]*)\s*([\d\.\+\-]*)\s*$/)
	{
		$Psi_lue[$test][0]=$1;
		$Psi_lue[$test][1]=$2;
		$Psi_lue[$test][2]=$3;
	}
	if($test>2 && /^\s*([\d\.\+\-]*)\s*$/)
	{
		$Energie_lue[$test-3]=$1;
	}
	$test++;
}

print $out "Fonction d'onde DDCI\n";
matrix_print(@Psi_lue)	;

#On corrige le dernier coefficient d'un facteur racine de deux car on a nu*(g-u+u-g) dans la fonction d'onde
for ($i=0; $i<3; $i++) 
{
	$Psi_lue[2][$i]*=sqrt(2);
}

#print $out "Fonction d'onde DDCI corrigée dernier coefficient\n";
#matrix_print(@Psi_lue)	;

#colonne_print(@Energie_lue);

#On shifte les énergie
my $E_trip=$Energie_lue[1];
for ($i=0; $i<4; $i++) 
{
	$Energie_lue[$i]-=$E_trip;
	$Energie_lue[$i]*=220000;
}

print $out "Énergies en cm-1 ramenées au triplet\n";
colonne_print(@Energie_lue);


#Calcul de la norme des vecteurs projetés, 
my @norme;
for ($i=0; $i<3; $i++) 
{
	$norme[$i]=$Psi_lue[0][$i]**2+$Psi_lue[1][$i]**2+$Psi_lue[2][$i]**2;
}

#print $out "Norme des trois fonctions d'ondes\n";
#colonne_print(@norme);

#On normalise la fonction d'onde Attention à la racine carrée
for ($i=0; $i<3; $i++) 
{
	for ($j=0; $j<3; $j++) 
	{
		$Psi_lue[$i][$j]/=sqrt($norme[$j]);
	}
}

#print $out "Psi DDCI normalisée dans l'espace modèle\n";
#matrix_print(@Psi_lue)	;


#On Calcule S
my $S=matrix_multiply(matrix_transpose(\@Psi_lue),\@Psi_lue);

print $out "Matrice de recouvrement\n";
matrix_print(@$S)	;



#déterminant de la matrice de recouvrement
my $det;
$det=-(-1+$S->[0][1]**2+$S->[0][2]**2+$S->[1][2]**2-2*$S->[0][1]*$S->[0][2]*$S->[1][2]);

#print $out "Déterminant de la matrice S\n";
#print $out $det."\n";

#Construction de l'inverse de la matrice
my @S_inv;
$S_inv[0][0]=(1-$S->[1][2]**2)/$det;
$S_inv[1][1]=(1-$S->[0][2]**2)/$det;
$S_inv[2][2]=(1-$S->[0][1]**2)/$det;
$S_inv[0][1]=(-$S->[0][1]+$S->[0][2]*$S->[1][2])/$det;
$S_inv[0][2]=(-$S->[0][2]+$S->[0][1]*$S->[1][2])/$det;
$S_inv[1][2]=(-$S->[1][2]+$S->[0][2]*$S->[0][1])/$det;
$S_inv[1][0]=$S_inv[0][1];
$S_inv[2][0]=$S_inv[0][2];
$S_inv[2][1]=$S_inv[1][2];
#	print $out "Matrice de recouvremenet inversée\n";
#matrix_print(@S_inv)	;

#définition de la matrice de passage de la passe g u à ionique neutre
my @P_ddci_ion;
for ($i=0; $i<3; $i++) 
{
	for ($j=0; $j<3; $j++) 
	{
		$P_ddci_ion[$i][$j]=0;
	}
}
$P_ddci_ion[0][0]=$P_ddci_ion[1][0]=$P_ddci_ion[1][1]=1/sqrt(2);
$P_ddci_ion[0][1]=-1/sqrt(2);
$P_ddci_ion[2][2]=1;


#Construction de la matrice Psi dans la base neutre,g/ionique,g/ionique/u
my $Psi=matrix_multiply(\@P_ddci_ion,\@Psi_lue);

print $out "Matrice remise dans la base ionique/neutre\n";
matrix_print(@$Psi)	;



#Construction de la matrice Psi_dague
#ICI il y a un problème dans la publi, la matrice psi_dague que je construit vérifie bien le fait que <psi_dague,i|psi_j>=delta_i,j
my $Psi_dague=matrix_multiply($Psi,\@S_inv);

#print $out "Matrice psi_dague non transposée\n";
#matrix_print(@$Psi_dague)	;


##############################################################"
#On transpose la matrice POur avoir psi dague  en tant que BRA
#################################################################
my $Psi_dague_transpo=matrix_transpose($Psi_dague);
#	print $out "Matrice psi_dague telle que définie dans la publi\n";
#matrix_print(@$Psi_dague_publi)	;


my $prod_scal=matrix_multiply($Psi_dague_transpo,$Psi);

#print $out "produit scalaire psi_dag*psi\n";
#matrix_print(@$prod_scal)	;

#Construction de l'hamiltonien effectif
#hamiltoniens à sommer

my @H_1;
for ($i=0; $i<3; $i++) 
{
	for ($j=0; $j<3; $j++) 
	{
		$H_1[$i][$j]=$Energie_lue[0]*$Psi->[$i][0]*$Psi_dague_transpo->[0][$j]; #
	}
}


my @H_2;
for ($i=0; $i<3; $i++) 
{
	for ($j=0; $j<3; $j++) 
	{
		$H_2[$i][$j]=$Energie_lue[2]*$Psi->[$i][1]*$Psi_dague_transpo->[1][$j]; # 
	}
}
my @H_3;
for ($i=0; $i<3; $i++) 
{
	for ($j=0; $j<3; $j++) 
	{
		$H_3[$i][$j]=$Energie_lue[3]*$Psi->[$i][2]*$Psi_dague_transpo->[2][$j];#
	}
}

#	print $out "H1\n";
#matrix_print(@H_1)	;
#	print $out "H2\n";
#matrix_print(@H_2)	;
#	print $out "H3\n";
#matrix_print(@H_3)	;

my @H_eff;
	$H_eff[0][0]=$H_1[0][0]+$H_2[0][0]+$H_3[0][0];
	$H_eff[0][1]=$H_1[0][1]+$H_2[0][1]+$H_3[0][1];
	$H_eff[1][0]=$H_1[1][0]+$H_2[1][0]+$H_3[1][0];
	$H_eff[1][1]=$H_1[1][1]+$H_2[1][1]+$H_3[1][1];	

	$H_eff[0][3]=$H_1[0][2]+$H_2[0][2]+$H_3[0][2];	
	$H_eff[1][3]=$H_1[1][2]+$H_2[1][2]+$H_3[1][2];	

	$H_eff[3][0]=$H_1[2][0]+$H_2[2][0]+$H_3[2][0];	
	$H_eff[3][1]=$H_1[2][1]+$H_2[2][1]+$H_3[2][1];
	$H_eff[3][3]=$H_1[2][2]+$H_2[2][2]+$H_3[2][2];
	$H_eff[2][0]=$H_eff[2][1]=$H_eff[2][2]=$H_eff[2][3]=$H_eff[0][2]=$H_eff[1][2]=$H_eff[3][2]=0;

#print $out "Hamiltonien effectif dans la base ionique/neutre\n";
#matrix_print(@H_eff)	;



#matrice de passage
my @P;
for ($i=0; $i<4; $i++) 
{
	for ($j=0; $j<4; $j++) 
	{
		$P[$i][$j]=0;
	}
}
$P[2][0]=$P[3][0]=$P[0][1]=$P[1][1]=$P[2][2]=$P[0][3]=1/sqrt(2);
$P[3][2]=$P[1][3]=-1/sqrt(2);
#matrix_print(@P)	;

#Inverse de la matrice de passage
my $P_inv=matrix_transpose(\@P);
#matrix_print(@P_inv)	;

#hamiltonien dans la base localisée H'=P*H_eff*P^-1
my $H_loc=matrix_multiply(\@P,matrix_multiply(\@H_eff,$P_inv));
print $out "Matrice dans la base localisée aa bb ab- ba-\n";
matrix_print(@$H_loc)	;

my $J;
my $K;
my $K_p;
my $t_in;
my $t_ni;
my $t_per;
my $t_m;
my $U;
$J=$Energie_lue[0];
$K=($H_loc->[0][1]+$H_loc->[1][0])/2;
my $k1;
$k1=($H_loc->[0][1]+$H_loc->[1][0]+$H_loc->[3][3])/3;
my $k2;
$k2=($H_loc->[0][1]+$H_loc->[1][0]+4*$H_loc->[3][3])/6;
my $k3;
$k3=($H_loc->[0][1]+$H_loc->[1][0]+2*$H_loc->[3][3])/4;
$t_m=($H_loc->[2][0]+$H_loc->[0][2]+$H_loc->[2][1]+$H_loc->[1][2])/4;
$U=($H_loc->[1][1]+$H_loc->[0][0])/2;


print $out "ATTENTION IL FAUT REDEFINIR PLUS PROPREMENT T U ET K!!!!! \n";
print $out "valeur de J en cm-1               : $J \n";
print $out "valeur moyenne de U en cm-1       : $U \n";
print $out "valeur moyenne de t en cm -1      : $t_m \n";
print $out "valeur de K moyenne sur 2         : $K \n";	
print $out "valeur de K moyenne sur 3 en cm-1 : $k1 \n";	
print $out "K moyenne sur 3 pondérée 6        : $k2 \n";
print $out "K moyenne sur 3 pondérée 4        : $k3 \n";	



